# Copyright (c) 2021 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

import logging

from github import Github, GithubException

from cgitize.repo import Repo


class GitHub:
    def __init__(self, access_token):
        self._impl = Github(access_token)

    def get_repo(self, repo):
        try:
            return self._impl.get_repo(repo.id)
        except GithubException:
            logging.error("Couldn't fetch repository: %s", repo.id)
            raise

    def get_user_repos(self, user):
        try:
            return self._impl.get_user(user.name).get_repos()
        except GithubException:
            logging.error("Couldn't fetch user repositories: %s", user.name)
            raise

    @staticmethod
    def convert_repo(repo, *args, **kwargs):
        return Repo.from_github(repo, *args, **kwargs)



# Copyright (c) 2021 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

import os

import tomli

from cgitize.bitbucket import Bitbucket
from cgitize.github import GitHub
from cgitize.gitlab import GitLab
from cgitize.repo import Repo


class Section:
    def __init__(self, impl):
        self.impl = impl

    def _get_config_value(self, key, required=True, default=None):
        if required and default is None:
            if not key in self.impl:
                raise RuntimeError(f'configuration value is missing: {key}')
        return self.impl.get(key, default)

    def _get_config_path(self, *args, **kwargs):
        return os.path.abspath(self._get_config_value(*args, **kwargs))

    def _get_config_or_env(self, key, env_name):
        val = self._get_config_value(key, required=False)
        if val is not None:
            return val
        if env_name in os.environ:
            return os.environ[env_name]
        return None


class MainSection(Section):
    DEFAULT_OUTPUT_DIR = '/var/tmp/cgitize'

    @property
    def output_dir(self):
        return self._get_config_path('output_dir', default=MainSection.DEFAULT_OUTPUT_DIR)

    @property
    def clone_url(self):
        return self._get_config_value('clone_url', required=False)

    @property
    def clone_via_ssh(self):
        return self._get_config_value('clone_via_ssh', default=True)

    @property
    def default_owner(self):
        return self._get_config_value('owner', required=False)


class GitHubSection(Section):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.users = UsersSection(self.impl.get('users', {}))
        self.repositories = RepositoriesSection(self.impl.get('repositories', {}))

    @property
    def access_token(self):
        return self._get_config_or_env('access_token', 'CGITIZE_GITHUB_ACCESS_TOKEN')

    @property
    def url_auth(self):
        return self.access_token


def two_part_url_auth(username, password):
        if username is None or password is None:
            return None
        return f'{username}:{password}'


class BitbucketSection(Section):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.users = UsersSection(self.impl.get('users', {}))
        self.repositories = RepositoriesSection(self.impl.get('repositories', {}))

    @property
    def app_password(self):
        return self._get_config_or_env('app_password', 'CGITIZE_BITBUCKET_APP_PASSWORD')

    @property
    def username(self):
        return self._get_config_or_env('username', 'CGITIZE_BITBUCKET_USERNAME')

    @property
    def url_auth(self):
        return two_part_url_auth(self.username, self.app_password)

    def enum_repositories(self):
        return map(HostedRepo, self.repositories.enum_repositories())


class GitLabSection(Section):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.users = UsersSection(self.impl.get('users', {}))
        self.repositories = RepositoriesSection(self.impl.get('repositories', {}))

    @property
    def access_token(self):
        return self._get_config_or_env('access_token', 'CGITIZE_GITLAB_ACCESS_TOKEN')

    @property
    def username(self):
        return self._get_config_or_env('username', 'CGITIZE_GITLAB_USERNAME')

    @property
    def url_auth(self):
        return two_part_url_auth(self.username, self.access_token)


class UsersSection(Section):
    def enum_users(self):
        return self.impl.values()


class RepositoriesSection(Section):
    def enum_repositories(self):
        return self.impl.values()


class User:
    def __init__(self, impl):
        if 'name' not in impl:
            raise ValueError("every user must have 'name'")
        self._impl = impl

    @property
    def name(self):
        return self._impl['name']

    @property
    def dir(self):
        return self._impl.get('dir')

    @property
    def skip(self):
        return self._impl.get('skip', [])


class HostedRepo:
    def __init__(self, impl):
        if 'id' not in impl:
            raise ValueError("every hosted repository must have 'id'")
        self._impl = impl

    @property
    def id(self):
        return self._impl['id']

    @property
    def dir(self):
        return self._impl.get('dir')


class Config:
    DEFAULT_PATH = '/etc/cgitize/cgitize.toml'

    @staticmethod
    def read(path):
        return Config(path)

    def __init__(self, path):
        self.path = os.path.abspath(path)
        with open(self.path, 'rb') as f:
            self.impl = tomli.load(f)
        self.main = MainSection(self.impl)
        self.repositories = RepositoriesSection(self.impl.get('repositories', {}))
        self.github = GitHubSection(self.impl.get('github', {}))
        self.bitbucket = BitbucketSection(self.impl.get('bitbucket', {}))
        self.gitlab = GitLabSection(self.impl.get('gitlab', {}))

    def _parse_explicit_repositories(self):
        for r in self.repositories.enum_repositories():
            yield Repo.from_config(r, self)

    def _parse_hosted_repositories(self, cfg, api):
        for r in cfg.repositories.enum_repositories():
            r = HostedRepo(r)
            yield api.convert_repo(api.get_repo(r), self, r.dir)
        for u in cfg.users.enum_users():
            u = User(u)
            for r in api.get_user_repos(u):
                r = api.convert_repo(r, self, u.dir)
                if r.name in u.skip:
                    continue
                yield r

    def _parse_github_repositories(self):
        github = GitHub(self.github.access_token)
        return self._parse_hosted_repositories(self.github, github)

    def _parse_bitbucket_repositories(self):
        bitbucket = Bitbucket(self.bitbucket.username, self.bitbucket.app_password)
        return self._parse_hosted_repositories(self.bitbucket, bitbucket)

    def _parse_gitlab_repositories(self):
        gitlab = GitLab(self.gitlab.access_token)
        return self._parse_hosted_repositories(self.gitlab, gitlab)

    def parse_repositories(self):
        yield from self._parse_explicit_repositories()
        yield from self._parse_github_repositories()
        yield from self._parse_bitbucket_repositories()
        yield from self._parse_gitlab_repositories()



# Copyright (c) 2018 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

import os.path

from cgitize.utils import url_remove_auth, url_replace_auth


class Repo:
    @staticmethod
    def from_config(src, config):
        if 'name' not in src:
            raise ValueError("every repository must have 'name'")
        name = src['name']
        desc = src.get('desc')
        homepage = src.get('homepage')
        owner = src.get('owner', config.main.default_owner)
        if 'clone_url' not in src:
            raise ValueError("every repository must have 'clone_url'")
        clone_url = src['clone_url']
        subdir = src.get('dir')
        return Repo(name, clone_url, owner=owner, desc=desc, homepage=homepage,
                    subdir=subdir)

    @staticmethod
    def from_github(src, config, subdir=None):
        name = src.name
        desc = src.description
        homepage = src.html_url
        owner = src.owner.name

        https_url = src.clone_url
        ssh_url = src.ssh_url
        clone_url = ssh_url if config.main.clone_via_ssh else https_url
        url_auth = None if config.main.clone_via_ssh else config.github.url_auth

        return Repo(name, clone_url, owner=owner, desc=desc, homepage=homepage,
                    url_auth=url_auth, subdir=subdir)

    @staticmethod
    def from_bitbucket(src, config, subdir=None):
        name = src['name']
        desc = src['description']
        homepage = src['links']['html']['href']
        owner = src['owner']['display_name']

        https_urls = [link for link in src['links']['clone'] if link['name'] == 'https']
        if len(https_urls) != 1:
            raise RuntimeError(f"no https:// clone URL for repository '{name}'?!")
        # Bitbucket leaves the username in the URL... Sigh.
        https_url = url_remove_auth(https_urls[0]['href'])

        ssh_urls = [link for link in src['links']['clone'] if link ['name'] == 'ssh']
        if len(ssh_urls) != 1:
            raise RuntimeError(f"no ssh:// clone URL for repository '{name}'?!")
        ssh_url = ssh_urls[0]['href']

        clone_url = ssh_url if config.main.clone_via_ssh else https_url
        url_auth = None if config.main.clone_via_ssh else config.bitbucket.url_auth

        return Repo(name, clone_url, owner=owner, desc=desc, homepage=homepage,
                    url_auth=url_auth, subdir=subdir)

    @staticmethod
    def from_gitlab(src, config, subdir=None):
        name = src.name
        desc = src.description
        homepage = src.web_url
        owner = src.namespace['name']

        https_url = src.http_url_to_repo
        ssh_url = src.ssh_url_to_repo
        clone_url = ssh_url if config.main.clone_via_ssh else https_url
        url_auth = None if config.main.clone_via_ssh else config.gitlab.url_auth

        return Repo(name, clone_url, owner=owner, desc=desc, homepage=homepage,
                    url_auth=url_auth, subdir=subdir)

    def __init__(self, name, clone_url, owner=None, desc=None, homepage=None,
                 url_auth=None, subdir=None):
        self._name = name
        self._desc = desc
        self._homepage = homepage
        self._owner = owner
        self._clone_url = clone_url
        self._url_auth = url_auth
        self._dir = subdir

    @property
    def name(self):
        return self._name

    @property
    def desc(self):
        if self._desc is not None and self._desc:
            return self._desc
        if self.homepage:
            return self.homepage
        if self.clone_url:
            return self.clone_url
        return self.name

    @property
    def homepage(self):
        return self._homepage

    @property
    def owner(self):
        return self._owner

    @property
    def clone_url(self):
        return self._clone_url

    @property
    def url_auth(self):
        return self._url_auth

    @property
    def clone_url_with_auth(self):
        if not self.url_auth:
            return self.clone_url
        return url_replace_auth(self.clone_url, self.url_auth)

    @property
    def dir(self):
        if self._dir is None:
            return self.name
        return os.path.join(self._dir, self.name)



# Copyright (c) 2018 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

from argparse import ArgumentParser
import logging
import sys

from cgitize.cgit import CGitRepositories, CGitServer
from cgitize.config import Config
from cgitize.utils import setup_logging


def parse_args(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    parser = ArgumentParser()
    parser.add_argument('--config', '-c', metavar='PATH',
                        default=Config.DEFAULT_PATH,
                        help='config file path')
    parser.add_argument('--repo', metavar='REPO_ID',
                        nargs='*', dest='repos',
                        help='repos to pull')
    parser.add_argument('--force', '-f', action='store_true',
                        help='overwrite existing repositories')
    parser.add_argument('--verbose', '-v', action='store_true',
                        help='verbose log output')
    return parser.parse_args(argv)


def main(args=None):
    args = parse_args()
    with setup_logging(args.verbose):
        config = Config.read(args.config)
        cgit_server = CGitServer(config.main.clone_url)
        output = CGitRepositories(config.main.output_dir, cgit_server, force=args.force)
        success = True
        for repo in config.parse_repositories():
            if args.repos is None or repo.name in args.repos:
                if not output.update(repo):
                    success = False
        if success:
            logging.info('All repositories were updated successfully')
            return 0
        else:
            logging.warning("Some repositories couldn't be updated!")
            return 1


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2018 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

from contextlib import contextmanager
import logging
import os
import stat
import subprocess
import sys
from urllib.parse import urlsplit, urlunsplit


@contextmanager
def setup_logging(verbose=False):
    level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=level,
        datefmt='%Y-%m-%d %H:%M:%S',
        format='%(asctime)s | %(levelname)s | %(message)s',
        # Log to stdout, because that's where subprocess's output goes (so that
        # the don't get interleaved).
        stream=sys.stdout)
    try:
        yield
    except Exception as e:
        logging.exception(e)
        raise


def run(*args, capture_output=False, **kwargs):
    stdout = None
    stderr = None
    if capture_output:
        stdout = subprocess.PIPE
        stderr = subprocess.STDOUT

    logging.debug('%s', args)
    result = subprocess.run(args, check=True, stdout=stdout, stderr=stderr,
                            encoding='utf-8', **kwargs)

    if result.stdout is not None:
        logging.debug('\n%s', result.stdout)
    return result.stdout


def try_run(*args, **kwargs):
    try:
        run(*args, **kwargs)
        return True
    except subprocess.CalledProcessError as e:
        return e.returncode == 0


def run_capture(*args, **kwargs):
    return run(*args, capture_output=True, **kwargs)


def try_run_capture(*args, **kwargs):
    try:
        return True, run(*args, capture_output=True, **kwargs)
    except subprocess.CalledProcessError as e:
        return e.returncode == 0, e.output


@contextmanager
def chdir(new_cwd):
    old_cwd = os.getcwd()
    os.chdir(new_cwd)
    try:
        yield
    finally:
        os.chdir(old_cwd)


@contextmanager
def protected_file(path):
    # 0600:
    new_permissions = stat.S_IRUSR | stat.S_IWUSR
    if os.path.exists(path):
        old_permissions = stat.S_IMODE(os.stat(path).st_mode)
        os.chmod(path, new_permissions)
        try:
            yield
        finally:
            os.chmod(path, old_permissions)
    else:
        with open(path, mode='w'):
            pass
        os.chmod(path, new_permissions)
        try:
            yield
        finally:
            os.unlink(path)


def url_replace_auth(url, username, password=None):
    parts = urlsplit(url)
    netloc = username
    if password is not None:
        netloc += f':{password}'
    netloc += f'@{parts.hostname}'
    if parts.port is not None:
        netloc += f':{parts.port}'
    parts = parts._replace(netloc=netloc)
    return urlunsplit(parts)


def url_remove_auth(url):
    parts = urlsplit(url)
    netloc = parts.hostname
    parts = parts._replace(netloc=netloc)
    return urlunsplit(parts)



# Copyright (c) 2021 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

import unittest

from github import Github, GithubException


class GitHubTests(unittest.TestCase):
    def setUp(self):
        self.github = Github()

    def test_nonexistent_repo(self):
        with self.assertRaises(GithubException):
            self.github.get_repo('doesnot/exist')

    def test_existing_repo(self):
        r = self.github.get_repo('egor-tensin/cgitize-test-repository')
        self.assertEqual(r.name, 'cgitize-test-repository')
        self.assertEqual(r.description, 'Test cgitize repository')
        self.assertEqual(r.owner.name, 'Egor Tensin')
        self.assertEqual(r.owner.login, 'egor-tensin')
        self.assertEqual(r.html_url, 'https://github.com/egor-tensin/cgitize-test-repository')
        self.assertEqual(r.clone_url, 'https://github.com/egor-tensin/cgitize-test-repository.git')
        self.assertEqual(r.ssh_url, 'git@github.com:egor-tensin/cgitize-test-repository.git')



# Copyright (c) 2021 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

import logging

from gitlab import Gitlab
from gitlab.exceptions import GitlabGetError

from cgitize.repo import Repo


class GitLab:
    def __init__(self, access_token):
        self._impl = Gitlab('https://gitlab.com', private_token=access_token)

    def get_repo(self, repo):
        try:
            return self._impl.projects.get(repo.id)
        except GitlabGetError:
            logging.error("Couldn't fetch repository: %s", repo.id)
            raise

    def get_user_repos(self, user):
        try:
            # Strictly speaking, GitLab supports the /users/:username/projects
            # endpoint, which means you shouldn't need to fetch the user first,
            # but I don't think python-gitlab 2.10.0 supports that?
            users = self._impl.users.list(username=user.name)
            if not users:
                raise RuntimeError(f"Couldn't find GitLab user: {user.name}")
            assert len(users) == 1
            user = users[0]
            return user.projects.list()
        except GitlabGetError:
            logging.error("Couldn't fetch user repositories: %s", user.name)
            raise

    @staticmethod
    def convert_repo(repo, *args, **kwargs):
        return Repo.from_gitlab(repo, *args, **kwargs)



# Copyright (c) 2021 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

from contextlib import contextmanager
import os

from cgitize import utils


GIT_ENV = os.environ.copy()
GIT_ENV['GIT_SSH_COMMAND'] = 'ssh -oBatchMode=yes -oLogLevel=QUIET -oStrictHostKeyChecking=no -oUserKnownHostsFile=/dev/null'


class Config:
    def __init__(self, path):
        self.path = path

    def exists(self):
        return os.path.exists(self.path)

    def open(self, mode='r'):
        return open(self.path, mode=mode, encoding='utf-8')

    def read(self):
        with self.open(mode='r') as fd:
            return fd.read()

    def write(self, contents):
        with self.open(mode='w') as fd:
            fd.write(contents)

    @contextmanager
    def backup(self):
        old_contents = self.read()
        try:
            yield old_contents
        finally:
            self.write(old_contents)

    # What follows is an exteremely loose interpretation of what the .gitconfig
    # syntax is. The source was git-config(1).

    class Section:
        def __init__(self, name, variables):
            Config.Section.validate_name(name)
            self.name = name
            self.variables = variables

        @staticmethod
        def validate_name(name):
            if not name:
                raise RuntimeError('section names cannot be empty')
            for c in name:
                if c.isalnum() or c == '-' or c == '.':
                    continue
                raise RuntimeError(f'section names must only contain alphanumeric characters, . or -: {name}')

        @staticmethod
        def format_name(name):
            return name

        def format(self):
            result = f'[{self.format_name(self.name)}]\n'
            result += ''.join((var.format() for var in self.variables))
            return result

    class Subsection:
        def __init__(self, section, name, variables):
            Config.Section.validate_name(section)
            Config.Subsection.validate_name(name)
            self.section = section
            self.name = name
            self.variables = variables

        @staticmethod
        def validate_name(name):
            if '\n' in name:
                raise RuntimeError(f'subsection names cannot contain newlines: {name}')

        def format_name(self):
            name = self.name
            # Escape the backslashes:
            name = name.replace('\\', '\\\\')
            # Escape the quotes:
            name = name.replace('"', '\\"')
            # Put in quotes:
            return f'"{name}"'

        def format(self):
            result = f'[{Config.Section.format_name(self.section)} {self.format_name()}]\n'
            result += ''.join((var.format() for var in self.variables))
            return result

    class Variable:
        def __init__(self, name, value):
            Config.Variable.validate_name(name)
            Config.Variable.validate_value(value)
            self.name = name
            self.value = value

        @staticmethod
        def validate_name(name):
            if not name:
                raise RuntimeError('variable names cannot be empty')
            for c in name:
                if c.isalnum() or c == '-':
                    continue
                raise RuntimeError(f'variable name can only contain alphanumeric characters or -: {name}')
            if not name[0].isalnum():
                raise RuntimeError(f'variable name must start with an alphanumeric character: {name}')

        @staticmethod
        def validate_value(value):
            pass

        def format_name(self):
            return self.name

        def format_value(self):
            value = self.value
            # Escape the backslashes:
            value = value.replace('\\', '\\\\')
            # Escape the supported escape sequences (\n, \t and \b):
            value = value.replace('\n', '\\n')
            value = value.replace('\t', '\\t')
            value = value.replace('\b', '\\b')
            # Escape the quotes:
            value = value.replace('"', '\\"')
            # Put in quotes:
            value = f'"{value}"'
            return value

        def format(self):
            return f'    {self.format_name()} = {self.format_value()}\n'


class Git:
    EXE = 'git'

    @staticmethod
    def check(*args, **kwargs):
        return utils.try_run(Git.EXE, *args, env=GIT_ENV, **kwargs)

    @staticmethod
    def capture(*args, **kwargs):
        return utils.try_run_capture(Git.EXE, *args, env=GIT_ENV, **kwargs)

    @staticmethod
    def get_global_config():
        return Config(os.path.expanduser('~/.gitconfig'))

    @staticmethod
    @contextmanager
    def setup_auth(repo):
        if not repo.url_auth:
            yield
            return
        config = Git.get_global_config()
        with utils.protected_file(config.path):
            with config.backup() as old_contents:
                variables = [Config.Variable('insteadOf', repo.clone_url)]
                subsection = Config.Subsection('url', repo.clone_url_with_auth, variables)
                new_contents = f'{old_contents}\n{subsection.format()}'
                config.write(new_contents)
                yield



# Copyright (c) 2021 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

import unittest

from atlassian.bitbucket.cloud import Cloud
from requests.exceptions import HTTPError


class BitbucketTests(unittest.TestCase):
    def setUp(self):
        self.bitbucket = Cloud(cloud=True)

    def test_nonexistent_repo(self):
        with self.assertRaises(HTTPError):
            self.bitbucket.repositories.get('doesnot/exist')

    def test_existing_repo(self):
        r = self.bitbucket.repositories.get('egor-tensin/cgitize-test-repository')
        self.assertEqual(r['name'], 'cgitize-test-repository')
        self.assertEqual(r['description'], 'Test cgitize repository')

        self.assertEqual(r['owner']['display_name'], 'Egor Tensin')
        self.assertEqual(r['owner']['nickname'], 'egor-tensin')

        self.assertEqual(r['links']['html']['href'], 'https://bitbucket.org/egor-tensin/cgitize-test-repository')

        clone_urls = [link for link in r['links']['clone'] if link['name'] == 'https']
        self.assertEqual(len(clone_urls), 1)
        self.assertEqual(clone_urls[0]['href'], 'https://bitbucket.org/egor-tensin/cgitize-test-repository.git')

        ssh_urls = [link for link in r['links']['clone'] if link['name'] == 'ssh']
        self.assertEqual(len(ssh_urls), 1)
        self.assertEqual(ssh_urls[0]['href'], 'git@bitbucket.org:egor-tensin/cgitize-test-repository.git')



# Copyright (c) 2021 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

import unittest

from gitlab import Gitlab
from gitlab.exceptions import GitlabGetError


class GitLabTests(unittest.TestCase):
    def setUp(self):
        self.gitlab = Gitlab('https://gitlab.com')

    def test_nonexistent_repo(self):
        with self.assertRaises(GitlabGetError):
            self.gitlab.projects.get('doesnot/exist')

    def test_existing_repo(self):
        r = self.gitlab.projects.get('egor-tensin/cgitize-test-repository')
        self.assertEqual(r.name, 'cgitize-test-repository')
        self.assertEqual(r.description, 'Test cgitize repository')
        self.assertEqual(r.namespace['name'], 'Egor Tensin')
        self.assertEqual(r.namespace['path'], 'egor-tensin')
        self.assertEqual(r.web_url, 'https://gitlab.com/egor-tensin/cgitize-test-repository')
        self.assertEqual(r.http_url_to_repo, 'https://gitlab.com/egor-tensin/cgitize-test-repository.git')
        self.assertEqual(r.ssh_url_to_repo, 'git@gitlab.com:egor-tensin/cgitize-test-repository.git')



# Copyright (c) 2021 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

import logging

from atlassian.bitbucket.cloud import Cloud
from requests.exceptions import HTTPError

from cgitize.repo import Repo


class Bitbucket:
    def __init__(self, username=None, password=None):
        self._impl = Cloud(username=username, password=password, cloud=True)

    def get_repo(self, repo):
        try:
            return self._impl.repositories.get(repo.id)
        except HTTPError:
            logging.error("Couldn't fetch repository: %s", repo.id)
            raise

    def get_user_repos(self, user):
        try:
            page = 1
            while True:
                params = {'page': page}
                response = self._impl.repositories.get(user.name, params=params)
                yield from iter(response['values'])
                if 'next' not in response:
                    break
                page += 1
        except HTTPError:
            logging.error("Couldn't fetch user repositories: %s", user.name)
            raise

    @staticmethod
    def convert_repo(repo, *args, **kwargs):
        return Repo.from_bitbucket(repo, *args, **kwargs)



# Copyright (c) 2018 Egor Tensin <Egor.Tensin@gmail.com>
# This file is part of the "cgitize" project.
# For details, see https://github.com/egor-tensin/cgitize.
# Distributed under the MIT License.

import logging
import os
import shutil

from cgitize.git import Git
from cgitize.utils import chdir


class CGitServer:
    def __init__(self, clone_url):
        self.clone_url = clone_url

    def get_clone_url(self, repo):
        if self.clone_url is None:
            return None
        return self.clone_url.format(repo=repo.dir)


class CGitRCWriter:
    def __init__(self, cgit_server):
        self.cgit_server = cgit_server

    @staticmethod
    def get_path(repo_dir):
        return os.path.join(repo_dir, 'cgitrc')

    def write(self, repo_dir, repo):
        with open(self.get_path(repo_dir), 'w') as fd:
            self._write_field(fd, 'clone-url', self._build_clone_url(repo))
            self._write_field(fd, 'owner', repo.owner)
            self._write_field(fd, 'desc', repo.desc)
            self._write_field(fd, 'homepage', repo.homepage)

    @staticmethod
    def _write_field(fd, field, value):
        if value is None:
            return
        fd.write(f'{field}={value}\n')

    def _build_clone_url(self, repo):
        clone_urls = []
        if repo.clone_url is not None:
            clone_urls.append(repo.clone_url)
        cgit_clone_url = self.cgit_server.get_clone_url(repo)
        if cgit_clone_url is not None:
            clone_urls.append(cgit_clone_url)
        if not clone_urls:
            return None
        clone_urls = ' '.join(clone_urls)
        return clone_urls


class AgeFile:
    @staticmethod
    def write(repo_dir):
        timestamp = AgeFile.get_age(repo_dir)
        if timestamp:
            os.makedirs(AgeFile.get_dir(repo_dir), exist_ok=True)
            with open(AgeFile.get_path(repo_dir), mode='w') as fd:
                fd.write(f'{timestamp}')

    @staticmethod
    def get_age(repo_dir):
        # https://git.zx2c4.com/cgit/tree/contrib/hooks/post-receive.agefile
        with chdir(repo_dir):
            success, output = Git.capture('for-each-ref', '--sort=-authordate', '--count=1', '--format=%(authordate:iso8601)')
            if not success:
                logging.error("Couldn't get the timestamp of the newest commit in repository: %s", repo_dir)
                return None
            return output

    @staticmethod
    def get_dir(repo_dir):
        return os.path.join(repo_dir, 'info', 'web')

    @staticmethod
    def get_path(repo_dir):
        return os.path.join(AgeFile.get_dir(repo_dir), 'last-modified')


class CGitRepositories:
    def __init__(self, output_dir, cgit_server, force=False):
        self.dir = self._make_dir(output_dir)
        self.cgitrc = CGitRCWriter(cgit_server)
        self.force = force

    @staticmethod
    def _make_dir(rel_path):
        abs_path = os.path.abspath(rel_path)
        os.makedirs(abs_path, exist_ok=True)
        return abs_path

    def get_repo_dir(self, repo):
        return os.path.join(self.dir, repo.dir)

    def update(self, repo):
        success = self._mirror_or_update(repo)
        if success:
            self.cgitrc.write(self.get_repo_dir(repo), repo)
            AgeFile.write(self.get_repo_dir(repo))
        return success

    def _mirror_or_update(self, repo):
        repo_dir = self.get_repo_dir(repo)

        if not os.path.isdir(repo_dir):
            # The local directory doesn't exist, mirror the new repository.
            return self._mirror(repo)

        with chdir(repo_dir):
            if not Git.check('rev-parse', '--is-inside-work-tree'):
                # Overwrite the existing directory if it's not a Git repository.
                logging.warning('Local directory is not a repository, going to overwrite it: %s', repo_dir)
                return self._mirror(repo)

            success, output = Git.capture('config', '--get', 'remote.origin.url')
            if not success:
                # Every repository managed by this script should have the
                # 'origin' remote. If it doesn't, it's trash. Overwrite the
                # existing directory, mirroring the repository in it.
                logging.warning("Local repository doesn't have remote 'origin', going to overwrite it: %s", repo_dir)
                return self._mirror(repo)

            if f'{repo.clone_url}\n' != output:
                # Jeez, there's a proper local repository in the target
                # directory already with a different upstream; something's
                # wrong, fix it manually.
                logging.warning("Existing repository '%s' doesn't match the specified clone URL: %s", repo.name, repo.clone_url)
                if self.force:
                    # Unless --force was specified, in which case we overwrite
                    # the repository.
                    return self._fix_upstream_url(repo) and self._update_existing(repo)
                return False

            # The local directory contains the local version of the upstream,
            # update it.
            return self._update_existing(repo)

    def _mirror(self, repo):
        logging.info("Mirroring repository '%s' from: %s", repo.name, repo.clone_url)
        repo_dir = self.get_repo_dir(repo)
        if os.path.isdir(repo_dir):
            try:
                shutil.rmtree(repo_dir)
            except Exception as e:
                logging.exception(e)
                return False
        with Git.setup_auth(repo):
            return Git.check('clone', '--mirror', '--quiet', repo.clone_url, repo_dir)

    def _fix_upstream_url(self, repo):
        repo_dir = self.get_repo_dir(repo)
        with chdir(repo_dir):
            return Git.check('remote', 'set-url', 'origin', repo.clone_url)

    def _update_existing(self, repo):
        logging.info("Updating repository '%s'", repo.name)
        repo_dir = self.get_repo_dir(repo)
        with chdir(repo_dir):
            with Git.setup_auth(repo):
                if not Git.check('remote', 'update', '--prune'):
                    return False
            # In case the local repository is not a bare repository, but a
            # full-fledged working copy:
            if Git.check('rev-parse', '--verify', '--quiet', 'origin/master'):
                return Git.check('reset', '--soft', 'origin/master')
            return True
